/*
 * $Id: starter_flash.c,v 1.6 2012/08/07 10:29:24 stm Exp $
 *
 * Starter Flash: Embed Flash in PDF
 *
 * Required software: PDFlib/PDFlib+PDI/PPS 9
 * Required data: Flash data files
 *
 * The Flash source files "mini.mxml" and "minivid.mxml" for the the Flash
 * video player "minivid.swf" and the Flash test application "mini.swf"
 * are included in the "data" directory.
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "pdflib.h"

/*
 * Utility function to create a button that calls an action on a rich-media
 * annotation.
 */
static
void
create_richmedia_action_button(PDF * const p,
        const int font, const int fontsize,
        const char * const caption,
        const char * const tooltip,
        const char * const target_name,
        const char * const functionname,
        const int arg_handle,
        const double xpos, const double ypos,
        const double width, const double height)
{
    int action;
    char optlist[256];
    char *endp = optlist;

    endp += sprintf(optlist, "target={%s} functionname=%s",
                    target_name, functionname);

    if (arg_handle != -1)
    {
        endp += sprintf(endp, " richmediaargs=%d", arg_handle);
    }

    action = PDF_create_action(p, "RichMediaExecute", optlist);

    sprintf(optlist, "caption={%s} "
                    "tooltip={%s} "
                    "action={up=%d} "
                    "bordercolor={gray 0} "
                    "font=%d fontsize=%d",
                    caption,
                    tooltip,
                    action,
                    font, fontsize);
    PDF_create_field(p,
            xpos, ypos, xpos + width, ypos + height,
            caption, 0, "pushbutton", optlist);
}

int
main(void)
{
    /* This is where the data files are. Adjust as necessary. */
    const char searchpath[] = "../data";
    const char outfile[] = "starter_flash.pdf";

    PDF *p = PDF_new();

    /* create a new PDFlib object */
    if (!p) {
        printf("Couldn't create PDFlib object (out of memory)!\n");
        return(2);
    }

    PDF_TRY(p) {
        const int pageheight = 600;
        const int pagewidth = 700;
        const int gap = pageheight / 20;

        const int fontsize = pageheight / 30;
        const int button_fontsize = (fontsize * 2) / 3;

        /* Layout for header */
        const int textbox_width = pagewidth - 2 * gap;
        const int textbox_height = fontsize;
        const int textbox_llx = gap;
        const int textbox_lly = pageheight - gap - fontsize;

        char optlist[1024];
        int font, undirty_action;

        sprintf(optlist, "SearchPath=%s", searchpath);
        PDF_set_option(p, optlist);


        /* This means we must check return values of load_font() etc. */
        PDF_set_option(p, "errorpolicy=return");

        if (PDF_begin_document(p, outfile, 0,
            "viewerpreferences={fitwindow=true} compatibility=1.7ext3") == -1) {
            printf("Error: %s\n", PDF_get_errmsg(p));
            PDF_delete(p);
            return(2);
        }

        PDF_set_info(p, "Creator", "PDFlib starter sample");
        PDF_set_info(p, "Title", "starter_flash");

        font = PDF_load_font(p, "Helvetica", 0, "winansi", "");
        if (font == -1) {
            printf("Error loading font: %s\n", PDF_get_errmsg(p));
            PDF_delete(p);
            return(2);
        }

        {
            /* Name of the flash video player application */
            const char video_player_file[] = "minivid.swf";

            /* Name of the video to play */
            const char video_file[] = "flashvid.flv";

            /* Layout for Flash area */
            const int flash_width = pagewidth - 2 * gap;
            const int flash_height = (flash_width * 10) / 18;
            const int flash_llx = textbox_llx;
            const int flash_lly = textbox_lly - gap - flash_height;

            /* Layout for pushbuttons */
            const int num_buttons = 5;
            const int button_gap = 20;
            const int button_width =
                (flash_width - (num_buttons - 1) * button_gap) / num_buttons;
            const int button_height = button_width / 2;
            const int button_llx = flash_llx;
            const int button_lly = flash_lly - gap - button_height;

#define NUM_CUEPOINTS 8
#define CUEPOINT_FIELDNAME "cuepoint"
            const int textfield_gap = button_gap;
            const int textfield_llx = flash_llx;
            const int textfield_height = button_height / 2;
            const int textfield_lly = button_lly - gap - textfield_height;
            const int textfield_width =
                (flash_width - (NUM_CUEPOINTS - 1) * textfield_gap)
                                                        / NUM_CUEPOINTS;
            const int textfield_fontsize = (button_fontsize * 3) / 4;

            int cuepoint_actions[NUM_CUEPOINTS];
            const int cuepoint_timestamps[NUM_CUEPOINTS] =
            {
                /*
                 * The cue point timestamps do not have any functionality,
                 * but are displayed in Acrobat in the properties of the
                 * Flash video, therefore they should be accurate.
                 */
                660, 1133, 1700, 2266, 2933, 3500, 3933, 4566
            };

            int video_player, video, vol20pct_args, vol100pct_args, button_xpos,
                i, textfield_xpos, rm_action, js_action;
            char *endp;

            /* Start a page */
            PDF_begin_page_ext(p, pagewidth, pageheight, "");

            /* Header line */
            sprintf(optlist, "position={center bottom} font=%d fontsize=%d "
                            "boxsize={%d %d}",
                            font, fontsize, textbox_width, textbox_height);
#define PLAYER_TITLE \
    "Video player with PDF button controls and cue point feedback"
            PDF_fit_textline(p, PLAYER_TITLE, 0, textbox_llx, textbox_lly,
                                optlist);
            PDF_create_bookmark(p, PLAYER_TITLE, 0, "");

            /* Load the video display flash application */
            video_player = PDF_load_asset(p, "Flash", video_player_file, 0, "");
            if (video_player == -1) {
                printf("Error loading video display application: %s\n",
                        PDF_get_errmsg(p));
                PDF_delete(p);
                return(2);
            }

            video = PDF_load_asset(p, "Video", video_file, 0, "");
            if (video == -1) {
                printf("Error loading video: %s\n", PDF_get_errmsg(p));
                PDF_delete(p);
                return(2);
            }

            /*
             * Create actions that are triggered by the cue points from the
             * Video. They change the visual appearance of the text fields at
             * the bottom of the page.
             */
            for (i = 0; i < NUM_CUEPOINTS; i += 1)
            {
                char javascript_buf[512];

                sprintf(javascript_buf,
                    "script {"
                        "var f = this.getField(\"" CUEPOINT_FIELDNAME "%d\"); "
                        "f.value = \"Cue point %d\"; "
                        "f.fillColor = color.blue;"
                    "}", i, i + 1);
                cuepoint_actions[i] = PDF_create_action(p, "JavaScript",
                                            javascript_buf);
            }

            /*
             * Build the option list for creating the richmedia annotation. The
             * name for the video player and video assets can be freely choosen,
             * but must be consistent between the "assets" declarations and
             * the "instances" declarations.
             *
             * The cue points must be matchable by name to the actual cue points
             * in the Video. The "time" options do not have any effect, but it
             * is recommended to add them with proper values as Acrobat
             * displays them in the properties dialog of the video.
             */
#define VIDEO_PLAYER_ASSET_NAME "minivid.swf"
#define VIDEO_ASSET_NAME "flashvid.flv"
#define VIDEO_PLAYER_NAME "Video Player"
            endp = optlist;
            endp += sprintf(endp,
                "name={" VIDEO_PLAYER_NAME "} "
                "richmedia={"
                    "activate={condition=opened} "
                    "assets={"
                        "{asset=%d name={" VIDEO_PLAYER_ASSET_NAME "}} "
                        "{asset=%d name={" VIDEO_ASSET_NAME "}} "
                    "} "
                    "configuration={"
                    "type=Video name=DefaultConfiguration "
                        "instances={"
                            "{"
                                "asset=" VIDEO_PLAYER_ASSET_NAME " "
                                "params={"
                                    "flashvars={"
                                        "source=" VIDEO_ASSET_NAME "&"
                                         "title=PDFlib%%20Video%%20Player"
                                    "} "
                                    "cuepoints={", video_player, video);
                                    for (i = 0; i < NUM_CUEPOINTS; i += 1)
                                    {
                                        endp += sprintf(endp,
                                                "{"
                                                    "type=event "
                                                    "name={Cue-Point %d} "
                                                    "time=%d "
                                                    "action={activate={%d}}"
                                                "} ",
                                                i + 1,
                                                cuepoint_timestamps[i],
                                                cuepoint_actions[i]);
                                    }
                                    strcpy(endp,
                                    "} "
                                "}"
                            "}"
                        "} "
                    "} "
                "}");

            /* Create the rich-media annotation */
            PDF_create_annotation(p, flash_llx, flash_lly,
                    flash_llx + flash_width, flash_lly + flash_height,
                    "RichMedia", optlist);

            /*
             * Create action buttons to drive the player
             */
            button_xpos = button_llx;
            create_richmedia_action_button(p,
                            font, button_fontsize,
                            "Play", "Play the video",
                            VIDEO_PLAYER_NAME, "play", -1,
                            button_xpos, button_lly,
                            button_width, button_height);

            button_xpos += button_gap + button_width;
            create_richmedia_action_button(p,
                            font, button_fontsize,
                            "Pause", "Pause the video",
                            VIDEO_PLAYER_NAME, "pause", -1,
                            button_xpos, button_lly,
                            button_width, button_height);

            /*
             * The rewind action is special because it not only executes a
             * RichMediaExecute action but also a JavaScript action to
             * reset the text fields for the cue point feedback.
             */
            button_xpos += button_gap + button_width;
            rm_action = PDF_create_action(p, "RichMediaExecute",
                    "target={" VIDEO_PLAYER_NAME "} functionname=rewind");
            sprintf(optlist,
                "script {"
                    "for (var i = 0; i < %d; i += 1)"
                    "{"
                        "var f = this.getField(\"" CUEPOINT_FIELDNAME
                                                                "\" + i); "
                        "f.value = \"\"; "
                        "f.fillColor = color.transparent;"
                    "}"
                "}", NUM_CUEPOINTS);
            js_action = PDF_create_action(p, "JavaScript", optlist);
            sprintf(optlist, "caption={Rewind} "
                            "tooltip={Rewind the video} "
                            "action={up={%d %d}} "
                            "bordercolor={gray 0} "
                            "font=%d fontsize=%d",
                            rm_action, js_action,
                            font, button_fontsize);
            PDF_create_field(p, button_xpos, button_lly,
                        button_xpos + button_width,
                        button_lly + button_height,
                        "Rewind", 0, "pushbutton", optlist);

            /*
             * Create an argument list containing one argument to set the
             * volume for the video player to 20%. Note that Acrobat expects
             * floating point numbers as PDF type string.
             */
            vol20pct_args =
                    PDF_poca_new(p, "containertype=array usage=richmediaargs");
            PDF_poca_insert(p, vol20pct_args,
                                    "type=string index=0 value={0.2}");
            button_xpos += button_gap + button_width;
            create_richmedia_action_button(p,
                            font, button_fontsize,
                            "Volume\n20%", "Set volume to 20%",
                            VIDEO_PLAYER_NAME, "volume", vol20pct_args,
                            button_xpos, button_lly,
                            button_width, button_height);

            /*
             * Create an argument list containing one argument to set the
             * volume for the video player to 100%.
             */
            vol100pct_args =
                PDF_poca_new(p, "containertype=array usage=richmediaargs");
            PDF_poca_insert(p, vol100pct_args,
                                "type=string index=0 value={1.0}");
            button_xpos += button_gap + button_width;
            create_richmedia_action_button(p,
                            font, button_fontsize,
                            "Volume\n100%", "Set volume to 100%",
                            VIDEO_PLAYER_NAME, "volume", vol100pct_args,
                            button_xpos, button_lly,
                            button_width, button_height);

            /*
             * Create textfields for giving feedback about the cue points that
             * are reached.
             */
            for (i = 0, textfield_xpos = textfield_llx; i < NUM_CUEPOINTS;
                    i += 1, textfield_xpos += textfield_gap + textfield_width)
            {
                char fieldname[128];

                sprintf(fieldname, CUEPOINT_FIELDNAME "%d", i);
                sprintf(optlist, "readonly=true font=%d fontsize=%d "
                            "fillcolor={gray 1} bordercolor={gray 0} "
                            "alignment=center",
                        font, textfield_fontsize);

                PDF_create_field(p, textfield_xpos, textfield_lly,
                        textfield_xpos + textfield_width,
                        textfield_lly + textfield_height,
                        fieldname, 0, "textfield", optlist);
            }

            PDF_end_page_ext(p, "");

            PDF_poca_delete(p, vol20pct_args, "");
            PDF_poca_delete(p, vol100pct_args, "");
        }

        {
            /* Name of the flash application */
            const char flash_app_file[] = "mini.swf";

            /* Layout for Flash area */
            const int flash_width = 400;
            const int flash_height = 300;
            const int flash_llx = pagewidth / 2 - flash_width / 2;
            const int flash_lly = textbox_lly - gap - flash_height;

            /* Layout for pushbuttons */
            const int button_gap = 20;
            const int button_width = flash_width / 5;
            const int button_height = button_width * 2 / 3;
            const int button_llx =
                    pagewidth / 2 - button_gap / 2 - button_width;
            const int button_lly = flash_lly - gap - button_height;

            int flash_app, button_xpos, flash_args, flash_args2;

            /* Start a page */
            PDF_begin_page_ext(p, pagewidth, pageheight, "");

            /* Header line */
            sprintf(optlist, "position={center bottom} font=%d fontsize=%d "
                            "boxsize={%d %d}",
                            font, fontsize, textbox_width, textbox_height);
#define PLAYER_TITLE2 "Passing parameters to functions in the Flash application"
            PDF_fit_textline(p, PLAYER_TITLE2, 0, textbox_llx, textbox_lly,
                                optlist);
            PDF_create_bookmark(p, PLAYER_TITLE2, 0, "");

            /* Load the flash application */
            flash_app = PDF_load_asset(p, "Flash", flash_app_file, 0, "");
            if (flash_app == -1) {
                printf("Error loading Flash application: %s\n",
                        PDF_get_errmsg(p));
                PDF_delete(p);
                return(2);
            }

            /*
             * Build the option list for creating the richmedia annotation.
             */
#define FLASH_APP_NAME "Flash Application"
            sprintf(optlist,
                "name={" FLASH_APP_NAME "} "
                "richmedia={"
                    "assets={"
                        "{asset=%d name={" FLASH_APP_NAME "}} "
                    "} "
                    "configuration={"
                        "type=Flash "
                        "instances={"
                            "{"
                                "asset={" FLASH_APP_NAME "} "
                                "params={"
                                    "binding=foreground "
                                    "flashvars={"
                                        "title=Click below for test"
                                        "&text="
                                    "}"
                                "}"
                            "}"
                        "} "
                    "} "
                    "activate={ "
                        "condition=visible "
                    "} "
                "}", flash_app);

            /* Create the rich-media annotation */
            PDF_create_annotation(p, flash_llx, flash_lly,
                    flash_llx + flash_width, flash_lly + flash_height,
                    "RichMedia", optlist);

            /*
             * Create argument list for usage with varargs-style function.
             * This only works as expected with strings and booleans.
             */
            flash_args =
                    PDF_poca_new(p, "containertype=array usage=richmediaargs");
            PDF_poca_insert(p, flash_args,
                    "type=string index=0 value={Got varargs parameters}");
            PDF_poca_insert(p, flash_args,
                    "type=string index=1 value={text string}");
            PDF_poca_insert(p, flash_args,
                    "type=integer index=2 value={42}");
            PDF_poca_insert(p, flash_args,
                    "type=float index=3 value={3.14159}");
            PDF_poca_insert(p, flash_args,
                    "type=boolean index=4 value=true");
            PDF_poca_insert(p, flash_args,
                    "type=string index=5 value={another string}");

            button_xpos = button_llx;
            create_richmedia_action_button(p,
                            font, button_fontsize,
                            "Varargs",
                            "Pass arguments to Flash application via "
                                    "variable-length argument list",
                            FLASH_APP_NAME, "myCallbackVarargs", flash_args,
                            button_xpos, button_lly,
                            button_width, button_height);

            /*
             * Create argument list for a function with types for the
             * parameters. Note that numbers must be passed as strings.
             */
            flash_args2 =
                    PDF_poca_new(p, "containertype=array usage=richmediaargs");
            PDF_poca_insert(p, flash_args2,
                    "type=string index=0 value={Got typed parameters}");
            PDF_poca_insert(p, flash_args2,
                    "type=boolean index=1 value=false");
            PDF_poca_insert(p, flash_args2,
                    "type=string index=2 value={3.14159}");
            PDF_poca_insert(p, flash_args2,
                    "type=string index=3 value={42}");
            PDF_poca_insert(p, flash_args2,
                    "type=string index=4 value={typed string argument}");

            button_xpos += button_gap + button_width;
            create_richmedia_action_button(p,
                            font, button_fontsize,
                            "Typed\narguments",
                            "Pass arguments to Flash application via typed "
                            "argument list",
                            FLASH_APP_NAME, "myCallbackTyped", flash_args2,
                            button_xpos, button_lly,
                            button_width, button_height);

            PDF_end_page_ext(p, "");

            PDF_poca_delete(p, flash_args, "");
            PDF_poca_delete(p, flash_args2, "");
        }

        /*
         * Prevent the prompt for saving the document, that is caused by
         * modifying the text fields for the cue points.
         */
        undirty_action = PDF_create_action(p, "JavaScript",
                            "script={ this.dirty = false; }");
        sprintf(optlist, "action={willclose=%d}", undirty_action);
        PDF_end_document(p, optlist);
    }
    PDF_CATCH(p) {
        printf("PDFlib exception occurred in starter_flash sample:\n");
        printf("[%d] %s: %s\n",
            PDF_get_errnum(p), PDF_get_apiname(p), PDF_get_errmsg(p));
        PDF_delete(p);
        return(2);
    }

    PDF_delete(p);
    return 0;
}
